# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import logging
import Queue
import re
import struct
import multiprocessing
import time

import bucky.names as names

from bucky.errors import ConfigError, ProtocolError
from bucky.metrics.counter import Counter
from bucky.metrics.gauge import Gauge
from bucky.metrics.histogram import Histogram
from bucky.metrics.meter import Meter
from bucky.metrics.timer import Timer
from bucky.udpserver import UDPServer


log = logging.getLogger(__name__)


class MetricsDCommand(object):
    UPDATE = object()
    CLEAR = object()
    DELETE = object()

    def __init__(self, name, mtype, action, value=None):
        self.name = name
        self.mtype = mtype
        self.action = action
        self.value = value
        if self.action is not self.UPDATE and self.value is not None:
            raise ValueError("Values are only valid for updates")


class MetricsDParser(object):
    NUMERIC_TYPES = {
        0x00: "!B", 0x01: "!b",
        0x10: "!H", 0x11: "!h",
        0x20: "!I", 0x21: "!i",
        0x30: "!Q", 0x31: "!q",
        0x40: "!f", 0x41: "!d"
    }

    METRIC_TYPES = {
        0x00: Counter,
        0x10: Gauge,
        0x20: Histogram,
        0x30: Meter,
        0x40: Timer
    }

    METRIC_ACTION = {
        0x00: MetricsDCommand.UPDATE,
        0x01: MetricsDCommand.CLEAR,
        0x02: MetricsDCommand.DELETE
    }

    def __init__(self):
        pass

    def parse(self, data):
        if data[0] != 0xAA:
            raise ProtocolError("Invalid magic byte")
        hostname, data = self.parse_string(data)
        while len(data):
            mc, data = self.parse_metric(hostname, data)
            yield mc

    def parse_metric(self, hostname, data):
        cmd, data = data[0], data[1:]
        mtype = cmd & 0xF0
        action = cmd & 0x0F
        if mtype not in self.METRIC_TYPES:
            raise ProtocolError("Invalid metric type")
        if action not in self.METRIC_ACTIONS:
            raise ProtocolError("Invalid metric action")
        name, data = self.parse_string(data)
        if action is MetricsDCommand.UPDATE:
            value, data = self.parse_number(data)
        else:
            value = None
        stat = names.statname(hostname, name.split("."))
        cmd = MetricsDCommand(stat, mtype, action, value)
        return cmd, data

    def parse_string(self, data):
        (length,) = struct.unpack("!H", data[:2])
        if length > len(data) - 2:
            raise ProtocolError("Truncated string value")
        if data[2+length] != 0x00:
            raise ProtocolError("String missing null-byte terminator")
        try:
            ret = data[2:2+length-1].decode("utf-8")
            return ret, data[2+length+1:]
        except UnicodeDecodeError:
            raise ProtocolError("String is not value UTF-8")

    def parse_number(self, data):
        fmt = self.NUMERIC_TYPES.get(data[0])
        if fmt is None:
            raise ProtocolError("Invalid numeric type")
        sz = struct.calcsize(fmt)
        if sz > len(data) - 1:
            raise ProtocolError("Truncated numeric value")
        (val,) = struct.unpack(data[1:1+sz])
        return val, data[1+sz:]


class MetricsDHandler(multiprocessing.Process):
    def __init__(self, outbox, interval):
        super(MetricsDHandler, self).__init__()
        self.daemon = True
        self.interval = interval
        self.outbox = outbox
        self.inbox = multiprocessing.Queue()
        self.next_update = time.time() + self.interval
        self.metrics = {}

    def enqueue(self, mc):
        self.inbox.put(mc)

    def update_metric(self, mc):
        if mc.action is MetricsDCommand.DELETE:
            metrics.pop(mc.name, None)
            return
        metric = self.metrics.get(mc.name)
        if mc.action is MetricsDCommand.CLEAR:
            if metric is not None:
                metric.clear()
            return
        assert mc.action is MetricsDCommand.UPDATE
        if metric is None:
            metric = mc.mtype(mc.name)
        metric.update(mc.value)

    def run(self):
        while True:
            to_sleep = self.next_update - time.time()
            if to_sleep <= 0:
                self.flush_updates()
            self.next_update = time.time() + self.interval
            to_sleep = self.interval
            try:
                mv = self.inbox.get(True, to_sleep)
                self.update_metric(mv)
            except Queue.Empty:
                continue

    def flush_updates(self):
        for _, metric in self.metrics.iteritems():
            for v in metric.metrics():
                self.outbox.put((v.name, v.value, v.time))


class MetricsDServer(UDPServer):
    def __init__(self, queue, cfg):
        super(MetricsDServer, self).__init__(cfg.metricsd_ip, cfg.metricsd_port)
        self.parser = MetricsDParser()
        self.handlers = self._init_handlers(queue, cfg)

    def handle(self, data, addr):
        try:
            for mc in self.parser.parse(data):
                handler = self._get_handler(mc.name)
                handler.enqueue(mc)
        except ProtocolError:
            log.exception("Error from: %s:%s" % addr)

    def _init_handlers(self, queue, cfg):
        ret = []
        default = cfg.metricsd_default_interval
        handlers = cfg.metricsd_handlers
        if not len(handlers):
            ret = [(None, MetricsDHandler(queue, default))]
            ret[0][1].start()
            return ret
        for item in handlers:
            if len(item) == 2:
                pattern, interval, priority = item[0], item[1], 100
            elif len(item) == 3:
                pattern, interval, priority = item
            else:
                raise ConfigError("Invalid handler specification: %s" % item)
            try:
                pattern = re.compile(pattern)
            except:
                raise ConfigError("Invalid pattern: %s" % pattern)
            if interval < 0:
                raise ConfigError("Invalid interval: %s" % interval)
            ret.append((pattern, interval, priority))
        handlers.sort(key=lambda p: p[2])
        ret = [(p, MetricsDHandler(queue, i)) for (p, i, _) in ret]
        ret.append((None, MetricsDHandler(queue, default)))
        for _, h in ret:
            h.start()
        return ret

    def _get_handler(self, name):
        for (p, h) in self.handlers:
            if p is None:
                return h
            if p.match(name):
                return h



import math

class EWMA(object):
    """\
    Exponentially-weighted moving avergage. Based on the
    implementation in Coda Hale's metrics library:
    
       https://github.com/codahale/metrics/blob/development/metrics-core/src/main/java/com/yammer/metrics/stats/EWMA.java
    """

    M1_ALPHA = 1 - math.exp(-5.0 / 60.0)
    M5_ALPHA = 1 - math.exp(-5.0 / 60.0 / 5.0)
    M15_ALPHA = 1 - math.exp(-5.0 / 60.0 / 15.0)

    @staticmethod
    def oneMinuteEWMA():
        return EWMA(EWMA.M1_ALPHA, 5.0)
    
    @staticmethod
    def fiveMinuteEWMA():
        return EWMA(EWMA.M5_ALPHA, 5.0)
    
    @staticmethod
    def fifteenMinuteEWMA():
        return EWMA(EWMA.M15_ALPHA, 5.0)

    def __init__(self, alpha, interval):
        self.alpha = alpha
        self.interval = interval
        self.curr_rate = None
        self.uncounted = 0L
    
    def update(self, val):
        self.uncounted += val
    
    def rate(self):
        return self.curr_rate
    
    def tick(self):
        count = self.uncounted
        self.uncounted = 0L
        instant_rate = count / self.interval
        if self.initialized:
            self.curr_rate += (self.alpha * (instant_rate - self.curr_rate))
        else:
            self.curr_rate = instant_rate





debug = False
log_level = "INFO"

metricsd_ip = "127.0.0.1"
metricsd_port = 23632
metricsd_enabled = True
metricsd_default_interval = 10.0
metricsd_handlers = []

collectd_ip = "127.0.0.1"
collectd_port = 25826
collectd_enabled = True
collectd_types = []
collectd_converters = []
collectd_use_entry_points = True

statsd_ip = "127.0.0.1"
statsd_port = 8125
statsd_enabled = True
statsd_flush_time = 10.0

graphite_ip = "127.0.0.1"
graphite_port = 2003
graphite_max_reconnects = 3
graphite_reconnect_delay = 5
graphite_pickle_enabled = False
graphite_pickle_buffer_size = 500

full_trace = False

name_prefix = None
name_postfix = None
name_replace_char = '_'
name_strip_duplicates = True
name_host_trim = []

custom_clients = []



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2011 Cloudant, Inc.

import heapq
import math
import random
import time

class ExpDecSample(object):
    """\
    An exponentially-decaying random sample of longs. Based
    on the implementation in Coda Hale's metrics library:
    
      https://github.com/codahale/metrics/blob/development/metrics-core/src/main/java/com/yammer/metrics/stats/ExponentiallyDecayingSample.java
    """
    
    RESCALE_THRESHOLD = 60 * 60 * 1000000000
    
    def __init__(self, reservoir_size, alpha):
        self.rsize = reservoir_size
        self.alpha = alpha
        self.values = []
        self.count = 0
        self.start_time = self.tick()
        self.next_rescale = self.start_time + self.RESCALE_THRESHOLD
    
    def clear(self):
        self.count = 0
        self.start_time = self.tick()
        self.next_rescale = self.start_time + self.RESCALE_THRESHOLD
    
    def size(self):
        return int(min(self.rsize, self.count))
    
    def update(self, val, when=None):
        if when is None:
            when = self.tick()
        priority = self.weight(when - self.start_time) / random.random()
        self.count += 1
        if self.count <= self.rsize:
            heapq.heappush(self.values, (priority, val))
        else:
            if priority > self.values[0][0]:
                heapq.heapreplace(self.values, (priority, val))
        now = self.tick()
        if now >= self.next_rescale:
            self.rescale(now, self.next_rescale)

    def rescale(self, now, next):
        # See the comment in the original Java implementation.
        self.next_rescale = now + self.RESCALE_THRESHOLD
        old_start = self.start_time
        self.start_time = self.tick()
        newvals = []
        factor = math.exp(-self.alpha * (self.start_time - old_start))
        for k, v in self.values:
            newvals.append((k * factor, v))
        self.values = newvals

    def tick(self):
        return long(time.time() * 1000000000.0)

    def weight(self, t):
        return math.exp(self.alpha * t)

    def get_values(self):
        return [v for (_, v) in self.values]



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2011 Cloudant, Inc.

import time

from bucky.metrics.metric import Metric, MetricValue as MV
from bucky.metrics.stats.ewma import EWMA

class Meter(Metric):
    def __init__(self, name):
        self.name = name
        self.count = 0
        self.m1_rate = EWMA.oneMinuteEWMA()
        self.m5_rate = EWMA.fiveMinuteEWMA()
        self.m15_rate = EWMA.fifteenMinuteEWMA()
        self.start_time = time.time()

    def update(self, value=1):
        self.count += value
        self.m1_rate.update(value)
        self.m5_rate.update(value)
        self.m15_rate.update(value)

    def metrics(self):
        for r in (self.m1_rate, self.m5_rate, self.m15_rate):
            r.tick()
        ret = []
        elapsed = time.time() - self.start_time
        ret.append(MV("%s.count" % self.name, self.count))
        ret.append(MV("%s.rate_avg" % self.name, float(self.count) / elapsed))
        ret.append(MV("%s.rate_1m" % self.name, self.m1_rate.rate()))
        ret.append(MV("%s.rate_5m" % self.name, self.m5_rate.rate()))
        ret.append(MV("%s.rate_15m" % self.name, self.m15_rate.rate()))
        return ret



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2012 Cloudant, Inc.

import multiprocessing

class Client(multiprocessing.Process):
    def __init__(self, pipe):
        super(Client, self).__init__()
        self.daemon = True
        self.pipe = pipe

    def run(self):
        while True:
            self.send(*self.pipe.recv())

    def send(self, host, name, value, time):
        raise NotImplemented()



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import logging
import socket
import sys
import multiprocessing

import bucky.cfg as cfg


log = logging.getLogger(__name__)


class UDPServer(multiprocessing.Process):
    def __init__(self, ip, port):
        super(UDPServer, self).__init__()
        self.daemon = True
        self.ip = ip
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.sock.bind((ip, port))
            log.info("Bound socket socket %s:%s" % (ip, port))
        except Exception:
            log.exception("Error binding socket %s:%s." % (ip, port))
            sys.exit(1)

    def run(self):
        while True:
            data, addr = self.sock.recvfrom(65535)
            log.debug("Received UDP packet from %s:%s" % addr)
            if data == 'EXIT':
                return
            if not self.handle(data, addr):
                return

    def handle(self, data, addr):
        raise NotImplemented()

    def close(self):
        self.send('EXIT')

    def send(self, data):
        log.debug("Sending UDP packet to %s:%s" % (self.ip, self.port))
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.sendto(data, (self.ip, self.port))






class BuckyError(Exception):
    def __init__(self, mesg):
        self.mesg = mesg
    def __str__(self):
        return self.mesg


class ConnectError(BuckyError):
    pass


class ConfigError(BuckyError):
    pass


class ProtocolError(BuckyError):
    pass



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2011 Cloudant, Inc.

import bucky.cfg as cfg


__host_trim__ = None


def _get_host_trim():
    global __host_trim__
    if __host_trim__ is not None:
        return __host_trim__
    host_trim = cfg.name_host_trim
    __host_trim__ = []
    for s in host_trim:
        s = list(reversed([p.strip() for p in s.split(".")]))
        __host_trim__.append(s)
    return __host_trim__


def hostname(host):
    host_trim = _get_host_trim()
    parts = host.split(".")
    parts = list(reversed([p.strip() for p in parts]))
    for s in host_trim:
        same = True
        for i, p in enumerate(s):
            if p != parts[i]:
                same = False
                break
        if same:
            parts = parts[len(s):]
            return parts
    return parts


def strip_duplicates(parts):
    ret = []
    for p in parts:
        if len(ret) == 0 or p != ret[-1]:
            ret.append(p)
    return ret


def statname(host, name):
    nameparts = name.split('.')
    parts = []
    if cfg.name_prefix:
        parts.append(cfg.name_prefix)
    if host:
        parts.extend(hostname(host))
    parts.extend(nameparts)
    if cfg.name_postfix:
        parts.append(cfg.name_postfix)
    if cfg.name_replace_char is not None:
        parts = [p.replace(".", cfg.name_replace_char) for p in parts]
    if cfg.name_strip_duplicates:
        parts = strip_duplicates(parts)
    return ".".join(parts)



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2011 Cloudant, Inc.

import math

from bucky.metrics.metric import Metric, MetricValue as MV
from bucky.metrics.stats.expdec_sample import ExpDecSample
from bucky.metrics.stats.usample import UniformSample


class Histogram(Metric):
    def __init__(self, name, biased=True, percentiles=None):
        self.name = name
        if biased:
            self.sample = ExpDecSample(1028, 0.015)
        else:
            self.sample = UniformSample(1028)
        self.percentiles = self._fmt(percentiles or (75, 85, 90, 95, 99, 99.9))
        self.count = 0
        self.sum = 0
        self.minv = None
        self.maxv = None
        self.variance = (-1.0, 0.0)

    def clear(self):
        self.sample.clear()
        self.count = 0
        self.sum = 0
        self.minv = None
        self.maxv = None
        self.variance = (-1.0, 0.0)

    def update(self, value):
        self.count += 1
        self.sum += value
        self.sample.update(value)
        if self.minv is None or value < self.minv:
            self.minv = value
        if self.maxv is None or value > self.maxv:
            self.maxv = value
        self._update_variance(value)

    def metrics(self):
        ret = []
        ret.append(MV("%s.count" % self.name, self.count))
        ret.append(MV("%s.sum" % self.name, self.sum))
        ret.append(MV("%s.min" % self.name, self.minv))
        ret.append(MV("%s.max" % self.name, self.maxv))
        if self.count > 0:
            ret.append(MV("%s.mean" % self.name, self.sum / self.count))
            ret.append(MV("%s.stddev" % self.name, self._stddev()))
            for disp, val in self._percentiles():
                name = "%s.%s" % (self.name, disp)
                ret.append(MV(name, val))
        return ret

    def _stddev(self):
        if self.count <= 1:
            return 0.0
        return math.sqrt(self.variance[1] / (float(self.count) - 1.0))

    def _update_variance(self, value):
        oldm, olds = self.variance
        if oldm == -1.0:
            self.variance = (value, 0.0)
            return
        newm = oldm + ((value - oldm) / self.count)
        news = olds + ((value - oldm) * (value - newm))
        self.variance = (newm, news)
    
    def _percentiles(self):
        values = self.sample.values()
        values.sort()
        ret = []
        for (p, d) in self.percentiles:
            pos = p * len(values + 1)
            if pos < 1:
                ret.append((d, values[0]))
            elif pos >= len(values):
                ret.append((d, values[-1]))
            else:
                lower, upper = values[int(pos-1)], values[int(pos)]
                percentile = lower + ((pos - math.floor(pos)) * (upper - lower))
                ret.append((d, percentile))
        return ret

    def _fmt(self, percentiles):
        ret = []
        for p in percentiles:
            d = "%0.1f" % p
            if d.endswith(".0"):
                d = p[:-2]
            d = "perc_%s" % d.replace(".", "_")
            ret.append((p, d))



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2011 Cloudant, Inc.

import logging
import socket
import struct
import sys
import time
try:
    import cPickle as pickle
except ImportError:
    import pickle

import bucky.client as client
import bucky.names as names


log = logging.getLogger(__name__)


class DebugSocket(object):
    def sendall(self, data):
        sys.stdout.write(data)

class CarbonClient(client.Client):
    def __init__(self, cfg, pipe):
        super(CarbonClient, self).__init__(pipe)
        self.debug = cfg.debug
        self.ip = cfg.graphite_ip
        self.port = cfg.graphite_port
        self.max_reconnects = cfg.graphite_max_reconnects
        self.reconnect_delay = cfg.graphite_reconnect_delay
        if self.max_reconnects <= 0:
            self.max_reconnects = sys.maxint
        self.connect()

    def connect(self):
        if self.debug:
            log.debug("Connected the debug socket.")
            self.sock = DebugSocket()
            return
        for i in xrange(self.max_reconnects):
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                self.sock.connect((self.ip, self.port))
                log.info("Connected to Carbon at %s:%s" % (self.ip, self.port))
                return
            except socket.error, e:
                if i+1 >= self.max_reconnects:
                    raise
                args = (self.ip, self.port, e)
                log.error("Failed to connect to %s:%s: %s" % args)
                if self.reconnect_delay > 0:
                    time.sleep(self.reconnect_delay)

    def reconnect(self):
        self.close()
        self.connect()

    def close(self):
        try:
            self.sock.close()
        except:
            pass

    def send(self, host, name, value, mtime):
        raise NotImplemented

class PlaintextClient(CarbonClient):
    def send(self, host, name, value, mtime):
        stat = names.statname(host, name)
        mesg = "%s %s %s\n" % (stat, value, mtime)
        for i in xrange(self.max_reconnects):
            try:
                self.sock.sendall(mesg)
                return
            except socket.error, err:
                if i+1 >= self.max_reconnects:
                    raise
                log.error("Failed to send data to Carbon server: %s" % err)
                self.reconnect()

class PickleClient(CarbonClient):
    def __init__(self, cfg, pipe):
        super(PickleClient, self).__init__(cfg, pipe)
        self.buffer_size = cfg.graphite_pickle_buffer_size
        self.buffer = []

    def send(self, host, name, value, mtime):
        stat = names.statname(host, name)
        self.buffer.append((stat, (mtime, value)))
        if len(self.buffer) >= self.buffer_size:
            self.transmit()

    def transmit(self):
        payload = pickle.dumps(self.buffer, protocol=-1)
        header = struct.pack("!L", len(payload))
        self.buffer = []
        for i in xrange(self.max_reconnects):
            try:
                self.sock.sendall(header + payload)
                return
            except socket.error, err:
                if i+1 >= self.max_reconnects:
                    raise
                log.error("Failed to send data to Carbon server: %s" % err)
                self.reconnect()




import random

class UniformSample(object):
    """\
    A random sample of a stream of long's based on the
    implementation in Coda Hale's Metrics library:
    
        https://github.com/codahale/metrics/blob/development/metrics-core/src/main/java/com/yammer/metrics/stats/UniformSample.java    
    """

    def __init__(self, size):
        self.count = 0
        self.values = [0] * size

    def clear(self):
        self.count = 0
        for i in range(len(self.values)):
            self.values[i] = 0

    def size(self):
        if self.count > len(self.values):
            return len(self.values)
        return self.count

    def update(self, val):
        self.count += 1
        if self.count <= len(self.values):
            self.values[self.count-1] = val
        else:
            r = random.random(0, self.count-1)
            if r < len(self.values):
                self.values[r] = val
    
    def get_values(self):
        return self.values[:]



# -*- coding: utf-8 -
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# Copyright 2011 Cloudant, Inc.

import logging
import multiprocessing
import optparse as op
import os
import Queue
import signal
import sys

import bucky
import bucky.cfg as cfg
import bucky.carbon as carbon
import bucky.collectd as collectd
import bucky.metricsd as metricsd
import bucky.statsd as statsd


log = logging.getLogger(__name__)


__usage__ = "%prog [OPTIONS] [CONFIG_FILE]"
__version__ = "bucky %s" % bucky.__version__

def options():
    return [
        op.make_option("--debug", dest="debug", default=False,
            action="store_true",
            help="Put server into debug mode. [%default]"
        ),
        op.make_option("--metricsd-ip", dest="metricsd_ip", metavar="IP",
            default=cfg.metricsd_ip,
            help="IP address to bind for the MetricsD UDP socket [%default]"
        ),
        op.make_option("--metricsd-port", dest="metricsd_port", metavar="INT",
            type="int", default=cfg.metricsd_port,
            help="Port to bind for the MetricsD UDP socket [%default]"
        ),
        op.make_option("--disable-metricsd", dest="metricsd_enabled",
            default=cfg.metricsd_enabled, action="store_false",
            help="Disable the MetricsD UDP server"
        ),
        op.make_option("--collectd-ip", dest="collectd_ip", metavar="IP",
            default=cfg.collectd_ip,
            help="IP address to bind for the CollectD UDP socket [%default]"
        ),
        op.make_option("--collectd-port", dest="collectd_port", metavar="INT",
            type='int', default=cfg.collectd_port,
            help="Port to bind for the CollectD UDP socket [%default]"
        ),
        op.make_option("--collectd-types", dest="collectd_types",
            metavar="FILE", action='append', default=cfg.collectd_types,
            help="Path to the collectd types.db file, can be specified multiple times"
        ),
        op.make_option("--disable-collectd", dest="collectd_enabled",
            default=cfg.collectd_enabled, action="store_false",
            help="Disable the CollectD UDP server"
        ),
        op.make_option("--statsd-ip", dest="statsd_ip", metavar="IP",
            default=cfg.statsd_ip,
            help="IP address to bind for the StatsD UDP socket [%default]"
        ),
        op.make_option("--statsd-port", dest="statsd_port", metavar="INT",
            type="int", default=cfg.statsd_port,
            help="Port to bind for the StatsD UDP socket [%default]"
        ),
        op.make_option("--disable-statsd", dest="statsd_enabled",
            default=cfg.statsd_enabled, action="store_false",
            help="Disable the StatsD server"
        ),
        op.make_option("--graphite-ip", dest="graphite_ip", metavar="IP",
            default=cfg.graphite_ip,
            help="IP address of the Graphite/Carbon server [%default]"
        ),
        op.make_option("--graphite-port", dest="graphite_port", metavar="INT",
            type="int", default=cfg.graphite_port,
            help="Port of the Graphite/Carbon server [%default]"
        ),
        op.make_option("--full-trace", dest="full_trace",
            default=cfg.full_trace, action="store_true",
            help="Display full error if config file fails to load"
        ),
        op.make_option("--log-level", dest="log_level",
            metavar="NAME", default="INFO",
            help="Logging output verbosity [%default]"
        ),
    ]


def main():
    parser = op.OptionParser(
        usage=__usage__,
        version=__version__,
        option_list=options()
    )
    opts, args = parser.parse_args()

    # Logging have to be configured before load_config,
    # where it can (and should) be already used
    logfmt = "[%(levelname)s] %(module)s - %(message)s"
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(logfmt))
    handler.setLevel(logging.ERROR) # Overridden by configuration
    logging.root.addHandler(handler)
    logging.root.setLevel(logging.DEBUG)

    if args:
        try:
            cfgfile, = args
        except ValueError:
            parser.error("Too many arguments.")
    else:
        cfgfile = None
    load_config(cfgfile, full_trace=opts.full_trace)

    if cfg.debug:
        cfg.log_level = "DEBUG"

    # Mandatory second commandline
    # processing pass to override values in cfg
    parser.parse_args(values=cfg)

    handler.setLevel(cfg.log_level)

    sampleq = multiprocessing.Queue()

    stypes = []
    if cfg.metricsd_enabled:
        stypes.append(metricsd.MetricsDServer)
    if cfg.collectd_enabled:
        stypes.append(collectd.CollectDServer)
    if cfg.statsd_enabled:
        stypes.append(statsd.StatsDServer)

    servers = []
    for stype in stypes:
        servers.append(stype(sampleq, cfg))
        servers[-1].start()

    if cfg.graphite_pickle_enabled:
        carbon_client = carbon.PickleClient
    else:
        carbon_client = carbon.PlaintextClient

    clients = []
    for client in cfg.custom_clients + [carbon_client]:
        send, recv = multiprocessing.Pipe()
        instance = client(cfg, recv)
        instance.start()
        clients.append((instance, send))

    def shutdown(signum, frame):
        for server in servers:
            server.close()
            sampleq.put(None)

    signal.signal(signal.SIGTERM, shutdown)

    while True:
        try:
            sample = sampleq.get(True, 1)
            if not sample:
                break
            for instance, pipe in clients:
                if not instance.is_alive():
                    log.error("Client process died. Exiting.")
                    sys.exit(1)
                pipe.send(sample)
        except Queue.Empty:
            pass
        for srv in servers:
            if not srv.is_alive():
                log.error("Server thread died. Exiting.")
                sys.exit(1)

    for child in multiprocessing.active_children():
        child.terminate()
        child.join()
    sys.exit()


def load_config(cfgfile, full_trace=False):
    cfg_mapping = vars(cfg)
    try:
        if cfgfile is not None:
            execfile(cfgfile, cfg_mapping)
    except Exception, e:
        log.error("Failed to read config file: %s" % cfgfile)
        if full_trace:
            log.exception("Reason: %s" % e)
        else:
            log.error("Reason: %s" % e)
        sys.exit(1)
    for name in dir(cfg):
        if name.startswith("_"):
            continue
        if name in cfg_mapping:
            setattr(cfg, name, cfg_mapping[name])


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass
    except Exception, e:
        raise
        print e
